
import pygame
from pygame.locals import KEYDOWN, K_ESCAPE

from pykitten import utils
from pykitten.gui import ClickArea, ImgButton, StaticWidget
from pykitten.engine import Engine

from game import Game, PacGame
from res import *
from basestates import BaseState, BgState
from map import Map, get_maps, load_map
from widgets import *

pygame.init()

class Editor(BaseState):
    def __init__(self, map=None, folder="custom"):
        BaseState.__init__(self)
        self.fps = 50
        pygame.mouse.set_visible(1)
        self.map = Map(map, edit=True, folder=folder)
        self.map_pos = (0,0)
        self.mapname = map
        self.folder = folder
        if map is None:
            self.mapname = "tmp.map"

        #generate grid layer
        max = self.map.max
        self.max = max
        width, height = max[0]*32, max[1]*32
        color = (255,255,255)
        grid = pygame.Surface((width, height))
        grid.set_alpha(100)
        for y in xrange(1, max[1]):
            pygame.draw.line(grid, color, (0, y*32), (width, y*32))
        for x in xrange(1, max[0]):
            pygame.draw.line(grid, color, (x*32, 0), (x*32, height))
        self.grid = grid

        #map clickeable area - rect values set in redraw()
        self.map_rect = ClickArea((0,0,1,1), callback=self.map_click)
        self.widgets.append(self.map_rect)

        #add tile buttons
        btns = ["ground", "catnip", "corner", "cataway", "ghost", "pacdog-img"]
        x, y = 0, 568
        for b in btns:
            btn = EditorButton(GFX[b], callback=self.select, name=b)
            btn.set_pos((x,y))
            self.add_button(b, btn)
            x += 35
        self.selected = "ground"
        self.select("ground")

        #add resize buttons
        btns = ["x+", "x-", "y+", "y-"]
        x, y = 580, 575
        for b in btns:
            btn = TxtButton(b, callback=self.resize_map, val=b)
            btn.set_pos((x,y))
            self.add_button(b, btn)
            x += 50

        #add new/save/load/test buttons
        x, y = 0, 0
        btn = TxtButton("New", callback=self.new)
        btn.set_pos((x,y))
        self.add_button("new", btn)

        x = x + btn.get_size()[0] + 50
        btn = TxtButton("Save", callback=self.save)
        btn.set_pos((x,y))
        self.add_button("save", btn)

        x = x + btn.get_size()[0] + 10
        btn = TxtButton("Load", callback=self.load)
        btn.set_pos((x,y))
        self.add_button("load", btn)

        x = x + btn.get_size()[0] + 10
        btn = TxtButton("Test", callback=self.test)
        btn.set_pos((x,y))
        self.add_button("test", btn)

        #add exit button
        btn = TxtButton("Exit", callback=self.quit)
        btn.set_pos((790-btn.get_size()[0], 0))
        self.add_button("exit", btn)

        self.upd = 1

    def unpause(self):
        leave = self.next and self.next.next
        nextnext = self.next.next
        BaseState.unpause(self)
        pygame.mouse.set_visible(1)
        self.upd = 1
        if leave:
            self.next = nextnext
            self.quit()


    def new(self):
        self.next = Editor(map=None, folder=self.folder)
        self.quit()

    def save(self):
        self.pause()
        self.next = SaveMenu(self.background, self.map, self.mapname,
            self.folder)
        self.quit()

    def test(self):
        self.map.save(self.mapname, folder=self.folder)
        self.pause()
        self.next = Game(mapname=self.mapname, folder=self.folder, test=True)
        self.quit()

    def load(self):
        self.pause()
        self.next = LoadMenu()
        self.quit()


    def resize_map(self, val):
        self.map.resize(val)
        self.upd = 1

    def map_click(self):
        x, y = pygame.mouse.get_pos()
        cx = (x-self.map_pos[0]) / 32
        cy = (y-self.map_pos[1]) / 32
        self.map.put_tile(self.selected, (cx, cy))
        self.upd = 1

    def update_other(self):
        if pygame.mouse.get_pressed()[0] and \
            self.map_rect.rect.collidepoint(pygame.mouse.get_pos()):
            self.map_click()

    def select(self, name):
        if name in self.buttons:
            self.buttons[self.selected].deselect()
            self.selected = name
            self.buttons[self.selected].select()
            self.upd = 1

    def redraw(self):
        #draw map
        self.background.fill((80,80,80))
        self.map.draw_items()
        self.map_pos = utils.center(self.map.get_size(),self.screen.get_size())
        self.background.blit(self.map, self.map_pos)

        #update map click rect
        self.map_rect.rect = pygame.Rect(self.map_pos, self.map.get_size())

        #draw grid
        grid = pygame.Surface(self.map.get_size())
        grid.set_colorkey((0,0,0))
        grid.blit(self.grid, (0,0))
        self.background.blit(grid, self.map_pos)

        #draw buttons
        for k in self.buttons.keys():
            self.buttons[k].draw(self.background)

    def update_screen(self):
        if self.upd:
            self.redraw()
            self.screen.blit(self.background, (0,0))
            pygame.display.flip()
            self.upd = 0



class SaveMenu(BgState):
    def __init__(self, img, map, mapname="", folder="custom"):
        BgState.__init__(self, img)
        self.map = map
        self.folder = folder
        pygame.mouse.set_visible(1)

        #background image + "save as"
        self.menu = StaticWidget((250,100))
        self.menu.fill((0,0,0))
        txt = TextLine("Save as:", fontsize=15)
        self.menu.blit(txt, (10,10))
        w, h = self.menu.get_size()
        pygame.draw.lines(self.menu, (0,0,255), True,
            ((0,0),(w-4,0),(w-4,h-4),(0,h-4)), 4)
        x, y = utils.center((w,h), (800,600))
        self.menu.set_pos((x,y))

        #text entry + ".map"
        self.entry = TextEntry(text=mapname[:-4])
        self.entry.set_pos((x+10, y + 30))
        txt = TextLine(".map", fontsize=20)
        self.menu.blit(txt, (self.entry.get_size()[0] + 15, 30))

        #save/cancel buttons
        msg_pos = self.menu.get_pos()
        size = self.menu.get_size()
        btn = FixedTextButton((100,24), "Cancel", callback=self.quit)
        btn.set_pos((msg_pos[0], msg_pos[1]+size[1]+4))
        btn.draw_border()
        self.add_button("cancel", btn)
        btn.draw(self.background)

        btn = FixedTextButton((100,24), "Save", callback=self.submit)
        btn.set_pos((msg_pos[0]+size[0]-100, msg_pos[1]+size[1]+4))
        btn.draw_border()
        self.add_button("save", btn)
        btn.draw(self.background)

        self.redraw()

        self.quit_timer = utils.Timer(20)
        self.quit_timer.deactivate()

        self.add_kbevent(KEYDOWN, K_ESCAPE, callback=self.quit)
        self.add_textentry_events(self.entry)

    def redraw(self):
        self.menu.draw(self.background)
        for b in self.buttons.keys():
            self.buttons[b].draw(self.background)
        self.entry.draw(self.background)
        self.upd = 1

    def update_other(self):
        if self.quit_timer.active:
            if self.quit_timer.update():
                self.quit()
        elif self.entry.update():
            self.redraw()

    def submit(self):
        if self.quit_timer.active:
            return

        filename = self.entry.text + ".map"
        if not self.map.save(filename, self.folder):
            txt = "Ok, map saved."
            self.next = Editor(filename, folder=self.folder)
        else:
            txt = "ERROR: Could not save map."

        size = self.menu.get_size()
        msg = pygame.Surface((size[0]-2, size[1]-2))
        msg.fill((0,0,0))
        txt = TextLine(txt)
        msg.blit(txt, utils.center(txt.get_size(), msg.get_size()))
        self.menu.blit(msg, (1,1))
        self.menu.draw(self.background)
        self.upd = 1

        self.quit_timer.activate()


class LoadMenu(BgState):
    def __init__(self, edit=True, folder="custom"):
        img = GFX["levelsel"]
        BgState.__init__(self, img)
        pygame.mouse.set_visible(1)
        self.preview = pygame.Surface((416,320))
        self.preview_pos = (312,105)
        self.edit = edit
        self.folder = folder
        self.stats = {"cats": 0, "dogs": 0, "catnip": 0, "cataway":0, "highscore":[]}
        self.scores = []

        #list of map filenames
        self.msg = TextLine("")
        self.maps = get_maps(self.folder)

        #buttons
        btn = ClickFlipButton(GFX["btn_up1"], GFX["btn_up2"],
            callback=self.set_page, add=-1)
        btn.set_pos((56,233))
        self.add_button("up", btn)

        btn = ClickFlipButton(GFX["btn_down1"], GFX["btn_down2"],
            callback=self.set_page, add=1)
        btn.set_pos((56,265))
        self.add_button("down", btn)

        btn = FixedTextButton((64,32), "Cancel", callback=self.quit)
        btn.set_pos((120,457))
        btn.draw_border()
        self.add_button("cancel", btn)

        btn = FixedTextButton((64,32), "Load", callback=self.load)
        btn.set_pos((216,457))
        btn.draw_border()
        self.add_button("load", btn)

        if not edit: #hide campaign maps in editor
            btn = ImgButton(GFX["sel_camp1"].copy(), callback=self.set_folder, val="lvl")
            btn.set_pos((117,41))
            self.add_button("lvl", btn)

        btn = ImgButton(GFX["sel_cus1"].copy(), callback=self.set_folder, val="custom")
        btn.set_pos((212,41))
        self.add_button("custom", btn)

        #pacmode unlocked when campaign done
        if SAVE["unlock"] >= 14:
            btn = ImgButton(GFX["mode_cat"], callback=self.set_pacmode)
            btn.set_pos((674,55))
            self.add_button("pacmode", btn)
        self.pacmode = False

        #stats/score area
        stats = StaticWidget((160, 96))
        stats.blit(GFX["ghost"], (3,0))
        stats.blit(GFX["pacdog-img"], (3,32))
        stats.blit(GFX["catnip"], (3, 64))
        stats.blit(GFX["cataway"], (96, 0))
        self.stat_bg = stats

        score_bg = StaticWidget((160, 96))
        self.score_bg = score_bg

        self.stat_flip = FlipImg([stats.copy(), score_bg.copy()])
        self.stat_flip.set_pos((472,457))

        #filename buttons
        size = (160, 30)
        file_pos = (120,106)
        x = file_pos[0]
        y = file_pos[1]
        for i in xrange(10):
            name = ""
            if i < len(self.maps):
                name = self.maps[i]
            btn = FixedTextButton(size, name,callback=self.select, nr=i)
            btn.set_pos((x,y))
            self.add_button(i, btn)
            y = y + size[1] + 2

        self.page = 0
        self.selected = 0
        self.max_page = 0
        self.set_folder(self.folder)
        self.select(0)

        self.add_kbevent(KEYDOWN, K_ESCAPE, self.quit)

    def set_folder(self, val):
        self.folder = val
        if val=="lvl":
            self.buttons["custom"].blit(GFX["sel_cus1"], (0,0))
            if not self.edit:
                self.buttons["lvl"].blit(GFX["sel_camp2"], (0,0))
            self.maps = get_maps(self.folder)[:SAVE["unlock"] + 1]
        else:
            self.buttons["custom"].blit(GFX["sel_cus2"], (0,0))
            if not self.edit:
                self.buttons["lvl"].blit(GFX["sel_camp1"], (0,0))
            self.maps = get_maps(self.folder)

        self.page = -1
        self.set_page(add=1)
        self.max_page = len(self.maps) / 11
        self.redraw()

    def load(self):
        if self.page * 10 + self.selected >= len(self.maps):
            return
        map = self.maps[self.page * 10 + self.selected]
        if self.edit:
            self.next = Editor(map, self.folder)
        else:
            self.pause()
            if self.pacmode:
                self.next = PacGame(map, folder=self.folder)
            else:
                self.next = Game(map, folder=self.folder, camp=False)
        self.quit()

    def unpause(self):
        BgState.unpause(self)
        self.select(self.selected)


    def set_page(self, add):
        p = self.page + add
        if 0 <= p <= self.max_page:
            for b in xrange(10):
                i = p * 10 + b
                btn = self.buttons[b]
                if i < len(self.maps):
                    btn.draw_text(self.maps[i])
                else:
                    btn.draw_text("")

            self.page += add
            self.select(0)
            self.redraw()

    def select(self, nr):
        if not self.buttons[nr].txt or nr >= len(self.maps):
            return

        self.buttons[self.selected].draw_border((0,0,0))
        self.selected = nr
        self.buttons[nr].draw_border((0,0,255))

        #generate preview
        mapname = self.maps[self.page * 10 + self.selected]
        map = Map(mapname, edit=True,
            folder=self.folder)
        map.draw_items()
        w, h = map.get_size()
        self.preview.fill((0,0,0))
        prev = pygame.transform.scale(map, (w/2, h/2))
        pos = utils.center(prev.get_size(), self.preview.get_size())
        self.preview.blit(prev, pos)

        #map stats
        self.stats["cats"] = len(map.ghost_start)
        self.stats["dogs"] = len(map.pac_start)
        self.stats["catnip"] = len(map.pellets)
        self.stats["cataway"] = len(map.cataway)

        tmp = (self.stats["cats"], self.stats["dogs"], self.stats["catnip"])
        img1 = self.stat_bg.copy()
        x, y = 64, 4
        for t in tmp:
            txt = TextLine(str(t))
            img1.blit(txt, (x,y))
            y += 32
        img1.blit(TextLine(str(self.stats["cataway"])), (128,4))

        #map highscore
        type = "custhigh"
        if self.folder == "lvl":
            type = "camphigh"
        if self.pacmode:
            type = "pachighca"
            if self.folder=="custom":
                type = "pachighcu"

        self.stats["highscore"] = []
        if SAVE[type].has_key(mapname):
            self.stats["highscore"] = SAVE[type][mapname]

        img2 = self.score_bg.copy()
        x, y = 0, 5
        tmp = self.stats["highscore"][:]
        tmp.reverse()
        for s in tmp:
            name, score = s
            txt = TextLine(name, fontsize=15)
            img2.blit(txt, (x,y))
            txt = TextLine(str(score), fontsize=15)
            img2.blit(txt, (x+120, y))
            y += 32

        if not self.stats["highscore"]:
            txt = TextLine("No highscores yet", 15)
            pos = utils.center(txt.get_size(), img2.get_size())
            img2.blit(txt, pos)
        self.stat_flip.set_images([img1, img2])

        self.redraw()

    def set_pacmode(self):
        self.pacmode = not self.pacmode
        if self.pacmode:
            self.buttons["pacmode"].blit(GFX["mode_dog"], (0,0))
        else:
            self.buttons["pacmode"].blit(GFX["mode_cat"], (0,0))
        self.select(self.selected)
        self.redraw()

    def redraw(self):
        self.background.blit(self.preview, self.preview_pos)
        self.stat_flip.draw(self.background)
        for b in self.buttons.keys():
            btn = self.buttons[b]
            btn.draw(self.background)
        self.upd = 1

    def update_other(self):
        self.stat_flip.update()
        self.redraw()


    def main_start(self):
        BgState.main_start(self)
        pygame.mouse.set_visible(1)


class EditorButton(HighlightButton):
    def __init__(self, img, callback, **kwargs):
        himg = img.copy()
        simg = img.copy()
        w, h = img.get_size()
        pygame.draw.lines(simg, (0,0,255), True,
            ((0,0),(w-2,0),(w-2,h-2),(0,h-2)), 2)
        HighlightButton.__init__(self, img, himg, simg, callback, **kwargs)




